# coding: utf-8

"""Utility methods for working with WSGI servers."""
from __future__ import print_function

import sys
import six
import webob.dec
import webob.exc
from oslo_config import cfg
from oslo_log import log as logging

__all__ = ["Application", "Middleware", "Debug"]

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
_ = (lambda v: v)


class Application(object):
    """Base WSGI application wrapper.

    Subclasses need to implement __call__.
    """

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.
        Any local configuration (that is, values under the [app:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.
        A hypothetical configuration would look like:
            [app:wadl]
            latest_version = 1.3
            paste.app_factory = nova.api.fancy_api:Wadl.factory
        which would result in a call to the `Wadl` class as
            import nova.api.fancy_api
            fancy_api.Wadl(latest_version='1.3')
        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.
        """
        return cls(**local_config)

    def __call__(self, environ, start_response):
        r"""Subclasses will probably want to implement __call__ like this:
        @webob.dec.wsgify(RequestClass=Request)
        def __call__(self, req):
          # Any of the following objects work as responses:
          # Option 1: simple string
          res = 'message\n'
          # Option 2: a nicely formatted HTTP exception page
          res = exc.HTTPForbidden(explanation='Nice try')
          # Option 3: a webob Response object (in case you need to play with
          # headers, or you want to be treated like an iterable, or or or)
          res = Response();
          res.app_iter = open('somefile')
          # Option 4: any wsgi app to be run next
          res = self.application
          # Option 5: you can get a Response object for a wsgi app, too, to
          # play with headers etc
          res = req.get_response(self.application)
          # You can then just return your response...
          return res
          # ... or set req.response and return None.
          req.response = res
        See the end of http://pythonpaste.org/webob/modules/dec.html
        for more info.
        """
        raise NotImplementedError(_('You must implement __call__'))


class Middleware(Application):
    """Base WSGI middleware.
    These classes require an application to be
    initialized that will be called next.  By default the middleware will
    simply call its wrapped app, or you can override __call__ to customize its
    behavior.
    """

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.
        Any local configuration (that is, values under the [filter:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.
        A hypothetical configuration would look like:
            [filter:analytics]
            redis_host = 127.0.0.1
            paste.filter_factory = nova.api.analytics:Analytics.factory
        which would result in a call to the `Analytics` class as
            import nova.api.analytics
            analytics.Analytics(app_from_paste, redis_host='127.0.0.1')
        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.
        """
        def _factory(app):
            return cls(app, **local_config)
        return _factory

    def __init__(self, application):
        self.application = application

    def process_request(self, req):
        """Called on each request.
        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.
        """
        return None

    def process_response(self, response):
        """Do whatever you'd like to the response."""
        return response

    @webob.dec.wsgify(RequestClass=webob.Request)
    def __call__(self, req):
        response = self.process_request(req)
        if response:
            return response
        response = req.get_response(self.application)
        return self.process_response(response)


class Debug(Middleware):
    """Helper class for debugging a WSGI application.
    Can be inserted into any WSGI application chain to get information
    about the request and response.
    """

    @webob.dec.wsgify(RequestClass=webob.Request)
    def __call__(self, req):
        print(('*' * 40) + ' REQUEST ENVIRON')
        for key, value in req.environ.items():
            print(key, '=', value)
        print()
        resp = req.get_response(self.application)

        print(('*' * 40) + ' RESPONSE HEADERS')
        for (key, value) in six.iteritems(resp.headers):
            print(key, '=', value)
        print()

        resp.app_iter = self.print_generator(resp.app_iter)

        return resp

    @staticmethod
    def print_generator(app_iter):
        """Iterator that prints the contents of a wrapper string."""
        print(('*' * 40) + ' BODY')
        for part in app_iter:
            sys.stdout.write(part)
            sys.stdout.flush()
            yield part
        print()


def get_app(app, middleware_cls, **kwargs):
    """A convenient function to wrap the middlewares."""

    def _get_cls(cls):
        if isinstance(cls, six.string_types):
            cls = __import__(cls)
        return cls

    for m in reversed(middleware_cls):
        cls = _get_cls(m)
        try:
            app = cls(app, kwargs)
        except TypeError:
            app = cls(app)

    return app



# coding: utf-8
"""
WSGI Server
    Run:
        python -m bservices.examples.example_server

HTTP Client:
    Add:
        curl http://127.0.0.1:10000/set_data -d '{"data": "test_data"}' \
             -H "Content-Type: application/json"
        >>> {"id": 1}
    Get:
        curl http://127.0.0.1:10000/get_data?id=1
        >>> {"data": "test_data", "id": 1}

Restriction of `bservices.wsgi.Resource`:
    (1) Only support three kinds of Content-Type:
        A. text/plain
        B. application/json
        C. application/xml
    (2) Automatically serialize and deserialize the body of the request,
        but only for application/json, text/plain. If you want to receive and
        handle the XML content, you MUST register the serializer and deserializer
        of XML; or, it will raise a exception.

Controller Action API:
    If using the wrapper of `Resource`, you obey the action rules below.
    Or, you MUST implement the engine to handle the HTTP request.

    The logic of `Resource` is simply.
    1. It gets the controller, the action method, the URL pattern argument
       by the result that `routes` parsed.
    2. deserialize the body, if any.
    3. pass all the arguments to the action method and call it.
    4. handle the result.

    Handle the result the action method returned:
    (1) if a dict, serialize it to JSON.
    (2) if a object of bservices.wsgi.ResponseObject, serialize it by default.
    (3) If any others, hand it to webob.dec.wsgify to handle.

    In the meanwhile, it also handle some convenient functions, such as the
    default status code, see `bservices.wsgi.Resource._process_stack` and
    `bservices.wsgi.response`. Moreover, it support the action mapping and the
    extensions.

    Action API:
        Args:
            (1) MUST have a argument, req, which is a object of
                `bservices.wsgi.Request`.
            (2) Arguments that `routes` parsed will be past to the action method.
            (3) If having a body, the action method MUST supply a argument, body.

        Return:
            (1) a dict object
            (2) a bservices.wsgi.ResponseObject object
            (3) a exception based on webob.exc.HTTPException or
                bservices.exception.ExceptionBase, or
                bservices.exception.ConvertedException.
            (4) a object of webob.exc.HTTPException, or
                bservices.exception.ConvertedException, or their subclass.
            (5) any object compatible with webob.dec.wsgify.__call__, for
                example, None, string(str, bytes, unicode), webob.Response, etc.

            Notice:
                The first two will be serialized.
                see bservices.wsgi.ResponseObject.

Middleware:
    Method One:
        >>> app = API()
        >>> app = Middleware3(app)
        >>> app = Middleware2(app)
        >>> app = Middleware1(app)

    Method Two:
        >>> from bservices.middleware import get_app
        >>> middlewares = ["PATH.TO.Middleware1", "PATH.TO.Middleware2",
                           "PATH.TO.Middleware3"]
        >>> app = API()
        >>> app = get_app(app, middlewares)

    def get_app(app, middlewares, **kwargs):
        `middlewares` is a list or tuple object, whose elements are a str or
        the Middleware class. Moreover, `kwargs` may pass the Middleware class
        as the second argument, which is a dict.

Decorators:
    bservices.wsgi carries some convenient decorators to implement some
    functions rapidly.

    These decorators are used on the action method of the controller.

    @serializers(**serializers)
        Assign the some serializers for the action method.

    @deserializers(**deserializers)
        Assign the some deserializers for the action method.

    @response(code)
        Assign the default status code for the action method.

    @action(name)
        Assign the action name for the action method.

    Notice:
        bservices.wsgi.Resource supplies two pairs of serializers and
        deserializers, that's, JSON and TEXT. The serializers of JSON and TEXT
        are JSONDeserializer and TextDeserializer, and the deserializers of JSON
        and TEXT are JSONDictSerializer and TextSerializer. Certianly, you can
        assign yourself serializers and deserializers to replace them.

        The serializers and the deserializers only serialize the body of the
        request, or deserialize the body of the response.

    For their usage examples, see below.
"""
import logging
import multiprocessing

import routes
import eventlet
from oslo_config import cfg
from oslo_log import log
from oslo_service import service
from oslo_service.wsgi import Router, Server

from bservices import wsgi, exception
from bservices.examples.db import api

LOG = logging.getLogger()
CONF = cfg.CONF

cli_opts = [
    cfg.StrOpt("listen_ip", default='0.0.0.0'),
    cfg.IntOpt("listen_port", default=10000)
]
CONF.register_cli_opts(cli_opts)


class DataController(wsgi.Controller):
    @wsgi.serializers(json=wsgi.JSONDictSerializer)
    @wsgi.deserializers(json=wsgi.JSONDeserializer)
    @wsgi.response(200)
    @wsgi.action("index")
    def get_data(self, req):
        try:
            id = int(req.GET["id"])
        except (KeyError, TypeError, ValueError):
            raise exception.BadRequest()

        ret = api.get_data(id)
        if not ret:
            raise exception.NotFound()
        return ret

    def set_data(self, req, body):
        try:
            data = body["data"]
        except (KeyError, TypeError, ValueError):
            raise exception.BadRequest()

        return api.set_data(data)


class API(Router):
    def __init__(self):
        mapper = routes.Mapper()
        mapper.redirect("", "/")

        resource = wsgi.Resource(DataController())
        mapper.connect("/get_data",
                       controller=resource,
                       action="get_data",
                       conditions={"method": ['GET']})
        mapper.connect("/set_data",
                       controller=resource,
                       action="set_data",
                       conditions={"method": ["POST"]})

        super(API, self).__init__(mapper)


class WSGIServer(Server, service.ServiceBase):
    pass


def main(project="example"):
    log.register_options(CONF)
    # log.set_defaults(default_log_levels=None)
    CONF(project=project)

    log.setup(CONF, project)
    eventlet.monkey_patch(all=True)

    server = WSGIServer(CONF, project, API(), host=CONF.listen_ip, port=CONF.listen_port,
                        use_ssl=False, max_url_len=1024)
    launcher = service.launch(CONF, server, workers=multiprocessing.cpu_count())
    launcher.wait()


if __name__ == '__main__':
    main()



# coding: utf-8
import sys

from oslo_config import cfg
from oslo_db import options as oslo_db_options
from oslo_db.sqlalchemy import utils as sqlalchemyutils
from oslo_log import log as logging

from bservices.db.sqlalchemy import get_engine, get_session

from . import models

CONF = cfg.CONF
CONF.register_opts(oslo_db_options.database_opts, 'database')
oslo_db_options.set_defaults(CONF, connection="sqlite:///:memory:")
DB_INIT = False

LOG = logging.getLogger(__name__)


def get_backend():
    """The backend is this module itself."""
    global DB_INIT
    if not DB_INIT:
        models.TestData.metadata.create_all(get_engine(CONF.database))
        DB_INIT = True
    return sys.modules[__name__]


###########################################################
# API Interface
def get_data(_id):
    model = models.TestData
    session = get_session(CONF.database)
    query = sqlalchemyutils.model_query(model, session)
    obj = query.filter_by(id=_id).first()
    if obj:
        return {
            "id": obj.id,
            "data": obj.data
        }
    else:
        return None


def set_data(data):
    model = models.TestData
    session = get_session(CONF.database)
    obj = model(data=data)
    obj.save(session)
    return {
        "id": obj.id,
    }



# coding: utf-8

from oslo_config import cfg
from oslo_db import concurrency
from oslo_log import log as logging

LOG = logging.getLogger(__name__)

CONF = cfg.CONF
_BACKEND_MAPPING = {'sqlalchemy': "bservices.examples.db.sqlalchemy.api"}

IMPL = concurrency.TpoolDbapiWrapper(CONF, backend_mapping=_BACKEND_MAPPING)


####################################
# API Interface
def get_data(id):
    return IMPL.get_data(id)


def set_data(data):
    return IMPL.set_data(data)



# encoding: utf-8
from __future__ import absolute_import, print_function, unicode_literals
import os
import socket
import logging
import functools
import traceback

import greenlet
import eventlet

from oslo_service import service, wsgi

LOG = logging.getLogger(__name__)


def listen_socket(host, port, backlog=1024, reuse=True):
    sock = socket.socket()
    if reuse:
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind((host, port))
    sock.listen(backlog)
    return sock


def wrap_exc(f):
    @functools.wraps(f)
    def inner(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception:
            return None
    return inner


def BaseRequestHandlerWrapper(HandlerClass):
    """A wrapper or adapter of BaseRequestHandler.

    By this, you can use BaseRequestHandler, StreamRequestHandler,
    DatagramRequestHandler of SocketServer in Python 2, or of socketserver in
    Python 3.
    """
    def wrapper(conn, addr):
        return HandlerClass(conn, addr, None)

    return wrapper


class HandlerBase(object):
    def __call__(self, client_sock, client_addr):
        raise NotImplementedError("__call__ MUST be implemented")


class ServerBase(service.ServiceBase):
    def __init__(self, pool_size=10240):
        self.pool_size = pool_size
        self._pool = eventlet.GreenPool(self.pool_size)
        self._server = None

    def serve(self, pool):
        raise NotImplementedError("The method of serve MUST be implemented")

    def _spawn(self, pool):
        pid = os.getpid()
        try:
            self.serve(pool)
        finally:
            pool.waitall()
            LOG.info("[Process{0}] the server exited".format(pid))

    def start(self):
        self._server = eventlet.spawn(self.serve, pool=self._pool)

    def stop(self):
        if self._server is not None:
            # let eventlet close socket
            self._pool.resize(0)
            self._server.kill()

    def wait(self):
        try:
            if self._server is not None:
                num = self._pool.running()
                LOG.debug("Waiting server to finish %d requests.", num)
                self._pool.waitall()
        except greenlet.GreenletExit:
            LOG.info("Server has stopped.")

    def reset(self):
        self._pool.resize(self.pool_size)


class TCPServer(ServerBase):
    def __init__(self, handler, host, port, pool_size=10240, backlog=1024, timeout=None):
        self.host = host
        self.port = port
        self.sock = listen_socket(self.host, self.port, backlog)
        LOG.info("Listen %s:%s" % (self.host, self.port))
        self.handler = handler
        self.timeout = timeout
        super(TCPServer, self).__init__(pool_size)

    def handle(self, conn, addr):
        try:
            self.handler(conn, addr)
        except socket.timeout:
            LOG.info("socket from {0} time out".format(addr))
        finally:
            try:
                conn.close()
            except socket.error:
                pass

    def serve(self, pool):
        pid = os.getpid()
        try:
            while True:
                try:
                    conn, addr = self.sock.accept()
                    conn.settimeout(self.timeout)
                    LOG.debug("[Process{0}] accepted {1}".format(pid, addr))
                    pool.spawn_n(self.handle, conn, addr)
                except socket.error as e:
                    LOG.error("[Process{0}] Socket has a error {1}: {2}".format(pid, addr, e))
                except Exception as e:
                    LOG.error(traceback.format_exc())
                    LOG.error("[Process{0}] can not handle the request from {1}: {2}".format(pid, addr, e))
                except (KeyboardInterrupt, SystemExit):
                    LOG.info("[Process{0}] the server is exiting".format(pid))
                    break
        finally:
            try:
                self.sock.close()
            except socket.error as e:
                pass
SocketServer = TCPServer


class TaskServer(ServerBase):
    EXTRA_TASH_NUM = 3

    def __init__(self, task_fn, *args, **kwargs):
        self.task_fn = task_fn
        self.task_num = kwargs.pop("task_num", 1)
        self.args = args
        self.kwargs = kwargs
        super(TaskServer, self).__init__(self.task_num + self.EXTRA_TASH_NUM)

    def _wrap_exc(self):
        try:
            self.task_fn(*self.args, **self.kwargs)
        except Exception:
            LOG.error(traceback.format_exc())

    def serve(self, pool):
        for i in range(self.task_num):
            pool.spawn_n(self._wrap_exc)
        LOG.info("Process{0} start {1} tasks".format(os.getpid(), self.task_num))


class PoolServer(ServerBase):
    def __init__(self, handler, *args, **kwargs):
        pool_size = kwargs.pop("pool_size", 10240)
        self.handler = handler
        self.args = args
        self.kwargs = kwargs
        super(PoolServer, self).__init__(pool_size)

    def serve(self, pool):
        try:
            self.handler(pool, *self.args, **self.kwargs)
        except Exception:
            LOG.error(traceback.format_exc())


# Below the version of 0.10.0, oslo_service.wsgi.Server doesn't inherit
# oslo_service.wsgi.ServiceBase. So, if using the version below 0.10.0,
# you SHOULD inherit oslo_service.wsgi.ServiceBase.
class WSGIServer(wsgi.Server):
    def __init__(self, *args, **kwargs):
        kwargs["pool_size"] = kwargs.get("pool_size", 10240)
        super(WSGIServer, self).__init__(*args, **kwargs)



# coding: utf-8

"""Base exception handling.
"""

from oslo_log import log as logging
import six
import webob.exc

_ = (lambda v: v)
LOG = logging.getLogger(__name__)
_exc_map = {}


def _exc_map_f():
    for http in dir(webob.exc):
        if http.startswith("HTTP"):
            cls = getattr(webob.exc, http)
            code = getattr(cls, "code", None)
            if code and issubclass(cls, webob.exc.WSGIHTTPException):
                _exc_map.setdefault(code, [])
                _exc_map[code].append(cls)
_exc_map_f()


def convert_exc(code, first=True, strict=False):
    clss = _exc_map.get(code, None)
    if clss is None:
        if strict:
            return None
        code = int(code // 100) * 100
        clss = _exc_map.get(code)

    if first:
        return clss[0]
    else:
        return clss


class ConvertedException(webob.exc.WSGIHTTPException):
    def __init__(self, code=0, title="", explanation=""):
        self.code = code
        self.title = title
        self.explanation = explanation
        super(ConvertedException, self).__init__()


def _cleanse_dict(original):
    """Strip all admin_password, new_pass, rescue_pass keys from a dict."""
    return {k: v for k, v in six.iteritems(original) if "_pass" not in k}


class ExceptionBase(Exception):
    """Base Nova Exception
    To correctly use this class, inherit from it and define
    a 'msg_fmt' property. That msg_fmt will get printf'd
    with the keyword arguments provided to the constructor.
    """
    msg_fmt = _("An unknown exception occurred.")
    code = 500
    headers = {}
    safe = False

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if 'code' not in self.kwargs:
            try:
                self.kwargs['code'] = self.code
            except AttributeError:
                pass

        if not message:
            try:
                message = self.msg_fmt % kwargs
            except Exception:
                LOG.exception(_('Exception in string format operation'))
                for name, value in six.iteritems(kwargs):
                    LOG.error("%s: %s" % (name, value))    # noqa

                message = self.msg_fmt

        self.message = message
        super(ExceptionBase, self).__init__(message)

    def format_message(self):
        # NOTE(mrodden): use the first argument to the python Exception object
        # which should be our full ExceptionBase message, (see __init__)
        return self.args[0]


class Forbidden(ExceptionBase):
    ec2_code = 'AuthFailure'
    msg_fmt = _("Not authorized.")
    code = 403


class AdminRequired(Forbidden):
    msg_fmt = _("User does not have admin privileges")


class Invalid(ExceptionBase):
    msg_fmt = _("Unacceptable parameters.")
    code = 400


class InvalidAttribute(Invalid):
    msg_fmt = _("Attribute not supported: %(attr)s")


class ServiceUnavailable(Invalid):
    msg_fmt = _("Service is unavailable at this time.")


class InvalidToken(Invalid):
    msg_fmt = _("The token '%(token)s' is invalid or has expired")


class InvalidConnectionInfo(Invalid):
    msg_fmt = _("Invalid Connection Info")


class InvalidHostname(Invalid):
    msg_fmt = _("Invalid characters in hostname '%(hostname)s'")


class InvalidContentType(Invalid):
    msg_fmt = _("Invalid content type %(content_type)s.")


class InvalidAPIVersionString(Invalid):
    msg_fmt = _("API Version String %(version)s is of invalid format. Must "
                "be of format MajorNum.MinorNum.")


class InvalidInput(Invalid):
    msg_fmt = _("Invalid input received: %(reason)s")


class VersionNotFoundForAPIMethod(Invalid):
    msg_fmt = _("API version %(version)s is not supported on this method.")


class BadRequest(ExceptionBase):
    msg_fmt = _("Bad Request")
    code = 400


class MalformedRequestBody(BadRequest):
    msg_fmt = _("Malformed message body: %(reason)s")


class NotFound(ExceptionBase):
    msg_fmt = _("Resource could not be found.")
    code = 404


class FileNotFound(NotFound):
    msg_fmt = _("File %(file_path)s could not be found.")


class ConfigNotFound(ExceptionBase):
    msg_fmt = _("Could not find config at %(path)s")


class PasteAppNotFound(ExceptionBase):
    msg_fmt = _("Could not load paste app '%(name)s' from %(path)s")


class CoreAPIMissing(ExceptionBase):
    msg_fmt = _("Core API extensions are missing: %(missing_apis)s")


# Cannot be templated, msg needs to be constructed when raised.
class InternalError(ExceptionBase):
    ec2_code = 'InternalError'
    msg_fmt = "%(err)s"


class SocketPortInUseException(ExceptionBase):
    msg_fmt = _("Not able to bind %(host)s:%(port)d, %(error)s")



# coding: utf-8
import threading
from oslo_db.sqlalchemy import session as db_session

_ENGINE_FACADE = {}
_LOCK = threading.Lock()


def _create_facade(conf_group):
    # This fragment is copied from oslo.db
    return db_session.EngineFacade(
        sql_connection=conf_group.connection,
        slave_connection=conf_group.slave_connection,
        sqlite_fk=False,
        autocommit=True,
        expire_on_commit=False,
        mysql_sql_mode=conf_group.mysql_sql_mode,
        idle_timeout=conf_group.idle_timeout,
        connection_debug=conf_group.connection_debug,
        max_pool_size=conf_group.max_pool_size,
        max_overflow=conf_group.max_overflow,
        pool_timeout=conf_group.pool_timeout,
        sqlite_synchronous=conf_group.sqlite_synchronous,
        connection_trace=conf_group.connection_trace,
        max_retries=conf_group.max_retries,
        retry_interval=conf_group.retry_interval)


def _create_facade_lazily(facade, conf_group):
    global _LOCK, _ENGINE_FACADE
    if _ENGINE_FACADE.get(facade) is None:
        with _LOCK:
            if _ENGINE_FACADE.get(facade) is None:
                _ENGINE_FACADE[facade] = _create_facade(conf_group)
    return _ENGINE_FACADE[facade]


def get_engine(conf_group, use_slave=False):
    """Get the Engine object.

    If `use_slave` is True, the operations of read and write are detached.
    """
    facade = _create_facade_lazily(conf_group._group.name, conf_group)
    return facade.get_engine(use_slave=use_slave)


def get_session(conf_group, use_slave=False, **kwargs):
    """Get the Session object.

    If `use_slave` is True, the operations of read and write are detached.
    """
    facade = _create_facade_lazily(conf_group._group.name, conf_group)
    return facade.get_session(use_slave=use_slave, **kwargs)



# encoding: utf-8
"""Turn the program into a daemon process."""


def daemon(umask=None, workdir=None, maxfd=None, keep_fds_open=False, stdio=False):
    """Turn the program into a daemon process.

    Detach this process from the controlling terminal and run it in the
    background as a daemon.
    """
    # Default daemon parameters.
    # File mode creation mask of the daemon.
    UMASK = umask if umask else 0o07

    # Default working directory for the daemon.
    WORKDIR = workdir if workdir else "/"

    # Default maximum for the number of available file descriptors.
    MAXFD = maxfd if maxfd else 1024

    # The standard I/O file descriptors are redirected to /dev/null by default.
    import os
    if (hasattr(os, "devnull")):
        REDIRECT_TO = os.devnull
    else:
        REDIRECT_TO = "/dev/null"

    try:
        # Fork a child process so the parent can exit.  This returns control to
        # the command-line or shell.  It also guarantees that the child will not
        # be a process group leader, since the child receives a new process ID
        # and inherits the parent's process group ID.  This step is required
        # to insure that the next call to os.setsid is successful.
        pid = os.fork()
    except OSError as e:
        raise Exception("%s [%d]" % (e.strerror, e.errno))

    if (pid == 0):  # The first child.
        # To become the session leader of this new session and the process group
        # leader of the new process group, we call os.setsid().  The process is
        # also guaranteed not to have a controlling terminal.
        os.setsid()

        # Is ignoring SIGHUP necessary?
        #
        # It's often suggested that the SIGHUP signal should be ignored before
        # the second fork to avoid premature termination of the process.  The
        # reason is that when the first child terminates, all processes, e.g.
        # the second child, in the orphaned group will be sent a SIGHUP.
        #
        # "However, as part of the session management system, there are exactly
        # two cases where SIGHUP is sent on the death of a process:
        #
        #   1) When the process that dies is the session leader of a session that
        #      is attached to a terminal device, SIGHUP is sent to all processes
        #      in the foreground process group of that terminal device.
        #   2) When the death of a process causes a process group to become
        #      orphaned, and one or more processes in the orphaned group are
        #      stopped, then SIGHUP and SIGCONT are sent to all members of the
        #      orphaned group." [2]
        #
        # The first case can be ignored since the child is guaranteed not to have
        # a controlling terminal.  The second case isn't so easy to dismiss.
        # The process group is orphaned when the first child terminates and
        # POSIX.1 requires that every STOPPED process in an orphaned process
        # group be sent a SIGHUP signal followed by a SIGCONT signal.  Since the
        # second child is not STOPPED though, we can safely forego ignoring the
        # SIGHUP signal.  In any case, there are no ill-effects if it is ignored.
        #
        import signal           # Set handlers for asynchronous events.
        signal.signal(signal.SIGHUP, signal.SIG_IGN)

        try:
            # Fork a second child and exit immediately to prevent zombies.  This
            # causes the second child process to be orphaned, making the init
            # process responsible for its cleanup.  And, since the first child is
            # a session leader without a controlling terminal, it's possible for
            # it to acquire one by opening a terminal in the future (System V-
            # based systems).  This second fork guarantees that the child is no
            # longer a session leader, preventing the daemon from ever acquiring
            # a controlling terminal.
            pid = os.fork()  # Fork a second child.
        except OSError as e:
            raise Exception("%s [%d]" % (e.strerror, e.errno))

        if (pid == 0):  # The second child.
            # Since the current working directory may be a mounted filesystem, we
            # avoid the issue of not being able to unmount the filesystem at
            # shutdown time by changing it to the root directory.
            os.chdir(WORKDIR)
            # We probably don't want the file mode creation mask inherited from
            # the parent, so we give the child complete control over permissions.
            os.umask(UMASK)
        else:
            # exit() or _exit()?  See below.
            os._exit(0)  # Exit parent (the first child) of the second child.
    else:
        # exit() or _exit()?
        # _exit is like exit(), but it doesn't call any functions registered
        # with atexit (and on_exit) or any registered signal handlers.  It also
        # closes any open file descriptors.  Using exit() may cause all stdio
        # streams to be flushed twice and any temporary files may be unexpectedly
        # removed.  It's therefore recommended that child branches of a fork()
        # and the parent branch(es) of a daemon use _exit().
        os._exit(0)  # Exit parent of the first child.

    # Close all open file descriptors.  This prevents the child from keeping
    # open any file descriptors inherited from the parent.  There is a variety
    # of methods to accomplish this task.  Three are listed below.
    #
    # Try the system configuration variable, SC_OPEN_MAX, to obtain the maximum
    # number of open file descriptors to close.  If it doesn't exists, use
    # the default value (configurable).
    #
    # try:
    #    maxfd = os.sysconf("SC_OPEN_MAX")
    # except (AttributeError, ValueError):
    #    maxfd = MAXFD
    #
    # OR
    #
    # if (os.sysconf_names.has_key("SC_OPEN_MAX")):
    #    maxfd = os.sysconf("SC_OPEN_MAX")
    # else:
    #    maxfd = MAXFD
    #
    # OR
    #
    # Use the getrlimit method to retrieve the maximum file descriptor number
    # that can be opened by this process.  If there is not limit on the
    # resource, use the default value.
    #
    import resource  # Resource usage information.
    maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
    if (maxfd == resource.RLIM_INFINITY):
        maxfd = MAXFD

    # Close only the standard file descriptors. This was added in order to
    # allow programs to daemonize themselfs after some files were opened.
    if keep_fds_open:
        maxfd = 2           # standard error (2)

    if stdio:
        start = 2
    else:
        start = 0

    # Iterate through and close all file descriptors.
    for fd in range(start, maxfd):
        try:
            os.close(fd)
        except OSError:  # ERROR, fd wasn't open to begin with (ignored)
            pass

    if not start:
        # Redirect the standard I/O file descriptors to the specified file.  Since
        # the daemon has no controlling terminal, most daemons redirect stdin,
        # stdout, and stderr to /dev/null.  This is done to prevent side-effects
        # from reads and writes to the standard I/O file descriptors.
        #
        # This call to open is guaranteed to return the lowest file descriptor,
        # which will be 0 (stdin), since it was closed above.
        os.open(REDIRECT_TO, os.O_RDWR)  # standard input (0)

        # Duplicate standard input to standard output and standard error.
        os.dup2(0, 1)           # standard output (1)
        os.dup2(0, 2)           # standard error (2)



# coding: utf-8
from __future__ import absolute_import, print_function

import six
import webob

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_service import wsgi

from . import exception

_ = (lambda v: v)
LOG = logging.getLogger(__name__)

_NOT_SERIALIZED_TYPES = [
    'text/plain'
]

_SUPPORTED_REQUEST_CONTENT_TYPES = [
    'text/plain',
    'application/json',
    'application/xml',
]

_SUPPORTED_RESPONSE_CONTENT_TYPES = [
    'application/json',
    'application/xml',
]

_MEDIA_TYPE_MAP = {
    'text/plain': 'text',
    'application/xml': 'xml',
    'application/json': 'json',
}

# These are typically automatically created by routes as either defaults
# collection or member methods.
_ROUTES_METHODS = [
    'create',
    'delete',
    'show',
    'update',
]

_METHODS_WITH_BODY = [
    'POST',
    'PUT',
]

# name of attribute to keep version method information
VER_METHOD_ATTR = 'versioned_methods'


def utf8(value):
    """Try to turn a string into utf-8 if possible."""
    if isinstance(value, six.text_type):
        return value.encode('utf-8')
    assert isinstance(value, str)
    return value


def get_supported_request_content_types():
    return _SUPPORTED_REQUEST_CONTENT_TYPES


def add_request_content_types(types):
    _SUPPORTED_REQUEST_CONTENT_TYPES.append(types)


def get_supported_response_content_types():
    return _SUPPORTED_RESPONSE_CONTENT_TYPES


def add_response_content_types(types):
    _SUPPORTED_RESPONSE_CONTENT_TYPES.append(types)


def get_media_map():
    return dict(_MEDIA_TYPE_MAP.items())


def add_media_map(maps):
    _MEDIA_TYPE_MAP.update(maps)


class Request(wsgi.Request):
    """Add some specific logic to the base webob.Request."""

    def __init__(self, *args, **kwargs):
        super(Request, self).__init__(*args, **kwargs)
        self._extension_data = {'db_items': {}}

    def cache_db_items(self, key, items, item_key='id'):
        """Allow API methods to store objects from a DB query to be
        used by API extensions within the same API request.

        An instance of this class only lives for the lifetime of a
        single API request, so there's no need to implement full
        cache management.
        """
        db_items = self._extension_data['db_items'].setdefault(key, {})
        for item in items:
            db_items[item[item_key]] = item

    def get_db_items(self, key):
        """Allow an API extension to get previously stored objects within
        the same API request.

        Note that the object data will be slightly stale.
        """
        return self._extension_data['db_items'][key]

    def get_db_item(self, key, item_key):
        """Allow an API extension to get a previously stored object
        within the same API request.

        Note that the object data will be slightly stale.
        """
        return self.get_db_items(key).get(item_key)

    @property
    def accept_content_type(self):
        accept = self.environ.get("wsgi.best_content_type", None)
        return accept or self.best_match_content_type()

    @accept_content_type.setter
    def accept_content_type(self, accept):
        self.environ['wsgi.best_content_type'] = accept

    def best_match_content_type(self):
        """Determine the requested response content-type."""
        if 'wsgi.best_content_type' not in self.environ:
            # Calculate the best MIME type
            content_type = None

            # Check URL path suffix
            parts = self.path.rsplit('.', 1)
            if len(parts) > 1:
                possible_type = 'application/' + parts[1]
                if possible_type in get_supported_response_content_types():
                    content_type = possible_type

            if not content_type:
                content_type = self.accept.best_match(get_supported_response_content_types())

            self.environ['wsgi.best_content_type'] = (content_type or 'application/json')

        return self.environ['wsgi.best_content_type']

    def get_content_type(self):
        """Determine content type of the request body.

        Does not do any body introspection, only checks header

        """
        if "Content-Type" not in self.headers:
            return None

        content_type = self.content_type

        # NOTE(markmc): text/plain is the default for eventlet and
        # other webservers which use mimetools.Message.gettype()
        # whereas twisted defaults to ''.
        # if not content_type or content_type == 'text/plain':
        if not content_type:
            return None

        if content_type not in get_supported_request_content_types():
            raise exception.InvalidContentType(content_type=content_type)

        return content_type

    def best_match_language(self):
        """Determine the best available language for the request.

        :returns: the best language match or None if the 'Accept-Language'
                  header was not available in the request.
        """
        if not self.accept_language:
            return None
        return self.accept_language.best_match("en_US")


class ActionDispatcher(object):
    """Maps method name to local methods through action name."""

    def dispatch(self, *args, **kwargs):
        """Find and call local method."""
        action = kwargs.pop('action', 'default')
        action_method = getattr(self, str(action), self.default)
        return action_method(*args, **kwargs)

    def default(self, data):
        raise NotImplementedError()


class TextDeserializer(ActionDispatcher):
    """Default request body deserialization."""

    def deserialize(self, datastring, action='default'):
        return self.dispatch(datastring, action=action)

    def default(self, datastring):
        return {"body": datastring}


class JSONDeserializer(TextDeserializer):

    def _from_json(self, datastring):
        try:
            return jsonutils.loads(datastring)
        except ValueError:
            msg = _("cannot understand JSON")
            raise exception.MalformedRequestBody(reason=msg)

    def default(self, datastring):
        return {'body': self._from_json(datastring)}


class TextSerializer(ActionDispatcher):
    def serialize(self, data, action="default"):
        return self.dispatch(data, action=action)

    def default(self, data):
        return str(data)


class DictSerializer(ActionDispatcher):
    """Default request body serialization."""

    def serialize(self, data, action='default'):
        return self.dispatch(data, action=action)

    def default(self, data):
        return jsonutils.dumps(data)


class JSONDictSerializer(DictSerializer):
    """Default JSON request body serialization."""
    pass


def serializers(**serializers):
    """Attaches serializers to a method.

    This decorator associates a dictionary of serializers with a
    method.  Note that the function attributes are directly
    manipulated; the method is not wrapped.
    """
    def decorator(func):
        if not hasattr(func, 'wsgi_serializers'):
            func.wsgi_serializers = {}
        func.wsgi_serializers.update(serializers)
        return func
    return decorator


def deserializers(**deserializers):
    """Attaches deserializers to a method.

    This decorator associates a dictionary of deserializers with a
    method.  Note that the function attributes are directly
    manipulated; the method is not wrapped.
    """
    def decorator(func):
        if not hasattr(func, 'wsgi_deserializers'):
            func.wsgi_deserializers = {}
        func.wsgi_deserializers.update(deserializers)
        return func
    return decorator


def response(code):
    """Attaches response code to a method.

    This decorator associates a response code with a method.  Note
    that the function attributes are directly manipulated; the method
    is not wrapped.
    """
    def decorator(func):
        func.wsgi_code = code
        return func
    return decorator


class ResponseObject(object):
    """Bundles a response object with appropriate serializers.

    Object that app methods may return in order to bind alternate
    serializers with a response object to be serialized.  Its use is
    optional.
    """

    def __init__(self, obj, code=None, headers=None, content_type=None,
                 **serializers):
        """Binds serializers with an object.

        Takes keyword arguments akin to the @serializer() decorator
        for specifying serializers.  Serializers specified will be
        given preference over default serializers or method-specific
        serializers on return.
        """
        self.obj = obj
        self.serializers = serializers
        self._default_code = 200
        self._code = code
        self._headers = headers or {}
        self.serializer = None
        self.media_type = None
        self.content_type = content_type

    def __getitem__(self, key):
        """Retrieves a header with the given name."""
        return self._headers[key.lower()]

    def __setitem__(self, key, value):
        """Sets a header with the given name to the given value."""
        self._headers[key.lower()] = value

    def __delitem__(self, key):
        """Deletes the header with the given name."""
        del self._headers[key.lower()]

    def _bind_method_serializers(self, meth_serializers):
        """Binds method serializers with the response object.

        Binds the method serializers with the response object.
        Serializers specified to the constructor will take precedence
        over serializers specified to this method.

        :param meth_serializers: A dictionary with keys mapping to
                                 response types and values containing
                                 serializer objects.
        """
        # We can't use update because that would be the wrong
        # precedence
        for mtype, serializer in meth_serializers.items():
            self.serializers.setdefault(mtype, serializer)

    def get_serializer(self, content_type, default_serializers=None):
        """Returns the serializer for the wrapped object.

        Returns the serializer for the wrapped object subject to the
        indicated content type.  If no serializer matching the content
        type is attached, an appropriate serializer drawn from the
        default serializers will be used.  If no appropriate
        serializer is available, raises InvalidContentType.
        """
        default_serializers = default_serializers or {}
        content_type = self.content_type or content_type

        try:
            mtype = get_media_map().get(content_type, content_type)
            if mtype in self.serializers:
                return mtype, self.serializers[mtype]
            else:
                return mtype, default_serializers[mtype]
        except (KeyError, TypeError):
            raise exception.InvalidContentType(content_type=content_type)

    def preserialize(self, content_type, default_serializers=None):
        """Prepares the serializer that will be used to serialize.

        Determines the serializer that will be used and prepares an
        instance of it for later call.  This allows the serializer to
        be accessed by extensions for, e.g., template extension.
        """
        mtype, serializer = self.get_serializer(content_type,
                                                default_serializers)
        self.media_type = mtype
        self.serializer = serializer()

    def serialize(self, request, content_type, default_serializers=None):
        """Serializes the wrapped object.

        Utility method for serializing the wrapped object.  Returns a
        webob.Response object.
        """
        content_type = self.content_type or content_type
        if self.serializer:
            serializer = self.serializer
        else:
            _mtype, _serializer = self.get_serializer(content_type,
                                                      default_serializers)
            serializer = _serializer()

        response = webob.Response()
        response.status_int = self.code
        for hdr, value in self._headers.items():
            response.headers[hdr] = utf8(str(value))
        response.headers['Content-Type'] = utf8(content_type)
        if self.obj is not None:
            response.body = serializer.serialize(self.obj)

        return response

    @property
    def code(self):
        """Retrieve the response status."""
        return self._code or self._default_code

    @property
    def headers(self):
        """Retrieve the headers."""
        return self._headers.copy()


def action_peek_json(body):
    """Determine action to invoke."""
    try:
        decoded = jsonutils.loads(body)
    except ValueError:
        msg = _("cannot understand JSON")
        raise exception.MalformedRequestBody(reason=msg)

    # Make sure there's exactly one key...
    if len(decoded) != 1:
        msg = _("too many body keys")
        raise exception.MalformedRequestBody(reason=msg)

    # Return the action and the decoded body...
    return decoded.keys()[0]


class ResourceExceptionHandler(object):
    """Context manager to handle Resource exceptions.

    Used when processing exceptions generated by API implementation
    methods (or their extensions).  Converts most exceptions to Fault
    exceptions, with the appropriate logging.
    """

    def __enter__(self):
        return None

    def __exit__(self, ex_type, ex_value, ex_traceback):
        if not ex_value:
            return True

        if isinstance(ex_value, exception.Forbidden):
            raise webob.exc.HTTPForbidden(explanation=ex_value.format_message())
        elif isinstance(ex_value, exception.VersionNotFoundForAPIMethod):
            raise
        elif isinstance(ex_value, exception.Invalid):
            raise exception.ConvertedException(code=ex_value.code,
                                               explanation=ex_value.format_message())
        elif isinstance(ex_value, TypeError):
            exc_info = (ex_type, ex_value, ex_traceback)
            LOG.error(_('Exception handling resource: %s'), ex_value,
                      exc_info=exc_info)
            raise webob.exc.HTTPBadRequest()
        elif isinstance(ex_value, webob.exc.HTTPException):
            LOG.info(_("HTTP exception thrown: %s"), ex_value)
            raise ex_value

        code = getattr(ex_value, "code", None)
        if code:
            raise exception.ConvertedException(code=code,
                                               explanation=ex_value.format_message())

        # We didn't handle the exception
        return False


class Resource(object):
    """WSGI app that handles (de)serialization and controller dispatch.

    WSGI app that reads routing information supplied by RoutesMiddleware
    and calls the requested action method upon its controller.  All
    controller action methods must accept a 'req' argument, which is the
    incoming wsgi.Request. If the operation is a PUT or POST, the controller
    method must also accept a 'body' argument (the deserialized request body).
    They may raise a webob.exc exception or return a dict, which will be
    serialized by requested content type.
    """

    support_api_request_version = False

    def __init__(self, controller, action_peek=None, inherits=None,
                 methods_with_body=None, **deserializers):
        """:param controller: object that implement methods created by routes lib
           :param action_peek: dictionary of routines for peeking into an
                               action request body to determine the
                               desired action
           :param inherits: another resource object that this resource should
                            inherit extensions from. Any action extensions that
                            are applied to the parent resource will also apply
                            to this resource.
        """
        self.controller = controller

        default_deserializers = dict(json=JSONDeserializer, text=TextDeserializer)
        default_deserializers.update(deserializers)

        self.default_deserializers = default_deserializers
        self.default_serializers = dict(json=JSONDictSerializer, text=TextSerializer)

        self.action_peek = dict(json=action_peek_json)
        self.action_peek.update(action_peek or {})

        # Copy over the actions dictionary
        self.wsgi_actions = {}
        if controller:
            self.register_actions(controller)

        # Save a mapping of extensions
        self.wsgi_extensions = {}
        self.wsgi_action_extensions = {}
        self.inherits = inherits
        self.methods_with_body = methods_with_body or _METHODS_WITH_BODY

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.
        Any local configuration (that is, values under the [app:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.
        A hypothetical configuration would look like:
            [app:wadl]
            latest_version = 1.3
            paste.app_factory = nova.api.fancy_api:Wadl.factory
        which would result in a call to the `Wadl` class as
            import nova.api.fancy_api
            fancy_api.Wadl(latest_version='1.3')
        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.
        """
        return cls(**local_config)

    def register_actions(self, controller):
        """Registers controller actions with this resource."""
        actions = getattr(controller, 'wsgi_actions', {})
        for key, method_name in actions.items():
            self.wsgi_actions[key] = getattr(controller, method_name)

    def register_extensions(self, controller):
        """Registers controller extensions with this resource."""
        extensions = getattr(controller, 'wsgi_extensions', [])
        for method_name, action_name in extensions:
            # Look up the extending method
            extension = getattr(controller, method_name)

            if action_name:
                # Extending an action...
                if action_name not in self.wsgi_action_extensions:
                    self.wsgi_action_extensions[action_name] = []
                self.wsgi_action_extensions[action_name].append(extension)
            else:
                # Extending a regular method
                if method_name not in self.wsgi_extensions:
                    self.wsgi_extensions[method_name] = []
                self.wsgi_extensions[method_name].append(extension)

    def get_action_args(self, request_environment):
        """Parse dictionary created by routes library."""
        # NOTE(Vek): Check for get_action_args() override in the
        # controller
        if hasattr(self.controller, 'get_action_args'):
            return self.controller.get_action_args(request_environment)

        try:
            args = request_environment['wsgiorg.routing_args'][1].copy()
        except (KeyError, IndexError, AttributeError):
            return {}

        args.pop('controller', None)
        args.pop('format', None)

        return args

    def get_body(self, request):
        try:
            content_type = request.get_content_type()
        except exception.InvalidContentType:
            LOG.debug("Unrecognized Content-Type provided in request")
            return None, ''

        return content_type, request.body

    def deserialize(self, meth, content_type, body):
        meth_deserializers = getattr(meth, 'wsgi_deserializers', {})
        try:
            mtype = get_media_map().get(content_type, content_type)
            if mtype in meth_deserializers:
                deserializer = meth_deserializers[mtype]
            else:
                deserializer = self.default_deserializers[mtype]
        except (KeyError, TypeError):
            raise exception.InvalidContentType(content_type=content_type)

        if (hasattr(deserializer, 'want_controller') and deserializer.want_controller):
            return deserializer(self.controller).deserialize(body)
        else:
            return deserializer().deserialize(body)

    def _should_have_body(self, request):
        return request.method in self.methods_with_body

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, request):
        """WSGI method that controls (de)serialization and method dispatch."""
        # Identify the action, its arguments, and the requested
        # content type
        action_args = self.get_action_args(request.environ)
        action = action_args.pop('action', None)
        content_type, body = self.get_body(request)
        accept = request.best_match_content_type()

        # NOTE(Vek): Splitting the function up this way allows for
        #            auditing by external tools that wrap the existing
        #            function.  If we try to audit __call__(), we can
        #            run into troubles due to the @webob.dec.wsgify()
        #            decorator.
        return self._process_stack(request, action, action_args,
                                   content_type, body, accept)

    def _process_stack(self, request, action, action_args,
                       content_type, body, accept):
        """Implement the processing stack."""
        # Get the implementing method
        try:
            meth, extensions = self.get_method(request, action,
                                               content_type, body)
        except (AttributeError, TypeError):
            return webob.exc.HTTPNotFound()
        except KeyError as ex:
            msg = _("There is no such action: {0}").format(ex.args[0])
            return webob.exc.HTTPBadRequest(explanation=msg)
        except exception.MalformedRequestBody:
            msg = _("Malformed request body")
            return webob.exc.HTTPBadRequest(explanation=msg)

        # Now, deserialize the request body...
        try:
            contents = {}
            if self._should_have_body(request):
                # allow empty body with PUT and POST
                if request.content_length == 0:
                    contents = {'body': None}
                else:
                    contents = self.deserialize(meth, content_type, body)
        except exception.InvalidContentType:
            msg = _("Unsupported Content-Type")
            return webob.exc.HTTPBadRequest(explanation=msg)
        except exception.MalformedRequestBody:
            msg = _("Malformed request body")
            return webob.exc.HTTPBadRequest(explanation=msg)

        # Update the action args
        action_args.update(contents)

        response = None
        try:
            with ResourceExceptionHandler():
                action_result = self.dispatch(meth, request, action_args)
        except webob.exc.HTTPException as ex:
            response = ex

        if not response:
            # No exceptions; convert action_result into a ResponseObject
            resp_obj = None
            if accept not in _NOT_SERIALIZED_TYPES:
                resp_obj = ResponseObject(action_result)
            elif isinstance(action_result, ResponseObject):
                resp_obj = action_result
            else:
                response = action_result

            # Run post-processing extensions
            if resp_obj:
                # Do a preserialize to set up the response object
                serializers = getattr(meth, 'wsgi_serializers', {})
                resp_obj._bind_method_serializers(serializers)
                if hasattr(meth, 'wsgi_code'):
                    resp_obj._default_code = meth.wsgi_code

                try:
                    resp_obj.preserialize(accept, self.default_serializers)

                    if not response:
                        response = resp_obj.serialize(request, accept,
                                                      self.default_serializers)
                except exception.InvalidContentType as err:
                    return webob.exc.HTTPBadRequest(explanation=err)

        if hasattr(response, 'headers'):
            for hdr, val in response.headers.items():
                # Headers must be utf-8 strings
                response.headers[hdr] = utf8(str(val))

        return response

    def get_method(self, request, action, content_type, body):
        meth, extensions = self._get_method(request,
                                            action,
                                            content_type,
                                            body)
        if self.inherits:
            _meth, parent_ext = self.inherits.get_method(request,
                                                         action,
                                                         content_type,
                                                         body)
            extensions.extend(parent_ext)
        return meth, extensions

    def _get_method(self, request, action, content_type, body):
        """Look up the action-specific method and its extensions."""
        # Look up the method
        try:
            if not self.controller:
                meth = getattr(self, action)
            else:
                meth = getattr(self.controller, action)
        except AttributeError:
            if (not self.wsgi_actions or
                action not in _ROUTES_METHODS + ['action']):
                # Propagate the error
                raise
        else:
            return meth, self.wsgi_extensions.get(action, [])

        if action == 'action':
            # OK, it's an action; figure out which action...
            mtype = get_media_map().get(content_type)
            action_name = self.action_peek[mtype](body)
        else:
            action_name = action

        # Look up the action method
        return (self.wsgi_actions[action_name],
                self.wsgi_action_extensions.get(action_name, []))

    def dispatch(self, method, request, action_args):
        """Dispatch a call to the action-specific method."""

        try:
            return method(req=request, **action_args)
        except exception.VersionNotFoundForAPIMethod:
            # We deliberately don't return any message information
            # about the exception to the user so it looks as if
            # the method is simply not implemented.
            return webob.exc.HTTPNotFound()


def action(name):
    """Mark a function as an action.

    The given name will be taken as the action key in the body.

    This is also overloaded to allow extensions to provide
    non-extending definitions of create and delete operations.
    """
    def decorator(func):
        func.wsgi_action = name
        return func
    return decorator


def extends(*args, **kwargs):
    """Indicate a function extends an operation.

    Can be used as either::

        @extends
        def index(...):
            pass

    or as::

        @extends(action='resize')
        def _action_resize(...):
            pass
    """
    def decorator(func):
        # Store enough information to find what we're extending
        func.wsgi_extends = (func.__name__, kwargs.get('action'))
        return func

    # If we have positional arguments, call the decorator
    if args:
        return decorator(*args)

    # OK, return the decorator instead
    return decorator


class ControllerMetaclass(type):
    """Controller metaclass.

    This metaclass automates the task of assembling a dictionary
    mapping action keys to method names.
    """

    def __new__(mcs, name, bases, cls_dict):
        """Adds the wsgi_actions dictionary to the class."""
        # Find all actions
        actions = {}
        extensions = []
        # start with wsgi actions from base classes
        for base in bases:
            actions.update(getattr(base, 'wsgi_actions', {}))

        for key, value in cls_dict.items():
            if not callable(value):
                continue
            if getattr(value, 'wsgi_action', None):
                actions[value.wsgi_action] = key
            elif getattr(value, 'wsgi_extends', None):
                extensions.append(value.wsgi_extends)

        # Add the actions and extensions to the class dict
        cls_dict['wsgi_actions'] = actions
        cls_dict['wsgi_extensions'] = extensions

        return super(ControllerMetaclass, mcs).__new__(mcs, name, bases, cls_dict)


@six.add_metaclass(ControllerMetaclass)
class Controller(object):
    """Default controller."""

    _view_builder_class = None

    def __init__(self, view_builder=None):
        """Initialize controller with a view builder instance."""
        if view_builder:
            self._view_builder = view_builder
        elif self._view_builder_class:
            self._view_builder = self._view_builder_class()
        else:
            self._view_builder = None



# coding: utf-8
from sqlalchemy import Column, String, Integer
from sqlalchemy.ext.declarative import declarative_base
from oslo_db.sqlalchemy import models

BASE = declarative_base()


class TestData(models.ModelBase, BASE):
    __tablename__ = 'test_data'

    id = Column(Integer, primary_key=True, autoincrement=True)
    data = Column(String(256), nullable=False)



# coding: utf-8

# Define the minimum and maximum version of the API across all of the
# REST API. The format of the version is:
# X.Y where:
#
# - X will only be changed if a significant backwards incompatible API
# change is made which affects the API as whole. That is, something
# that is only very very rarely incremented.
#
# - Y when you make any change to the API. Note that this includes
# semantic changes which may not affect the input or output formats or
# even originate in the API code layer. We are not distinguishing
# between backwards compatible and backwards incompatible changes in
# the versioning system. It must be made clear in the documentation as
# to what is a backwards compatible change and what is a backwards
# incompatible one.

import re

from . import exception
_ = (lambda v: v)


class APIVersionRequest(object):
    """This class represents an API Version Request with convenience
    methods for manipulation and comparison of version
    numbers that we need to do to implement microversions.
    """

    def __init__(self, version_string=None):
        """Create an API version request object.

        :param version_string: String representation of APIVersionRequest.
            Correct format is 'X.Y', where 'X' and 'Y' are int values.
            None value should be used to create Null APIVersionRequest,
            which is equal to 0.0
        """
        self.ver_major = 0
        self.ver_minor = 0

        if version_string is not None:
            match = re.match(r"^([1-9]\d*)\.([1-9]\d*|0)$",
                             version_string)
            if match:
                self.ver_major = int(match.group(1))
                self.ver_minor = int(match.group(2))
            else:
                raise exception.InvalidAPIVersionString(version=version_string)

    def __str__(self):
        """Debug/Logging representation of object."""
        return ("API Version Request Major: %s, Minor: %s"
                % (self.ver_major, self.ver_minor))

    def is_null(self):
        return self.ver_major == 0 and self.ver_minor == 0

    def _format_type_error(self, other):
        return TypeError(_("'%(other)s' should be an instance of '%(cls)s'") %
                         {"other": other, "cls": self.__class__})

    def __lt__(self, other):
        if not isinstance(other, APIVersionRequest):
            raise self._format_type_error(other)

        return ((self.ver_major, self.ver_minor) <
                (other.ver_major, other.ver_minor))

    def __eq__(self, other):
        if not isinstance(other, APIVersionRequest):
            raise self._format_type_error(other)

        return ((self.ver_major, self.ver_minor) ==
                (other.ver_major, other.ver_minor))

    def __gt__(self, other):
        if not isinstance(other, APIVersionRequest):
            raise self._format_type_error(other)

        return ((self.ver_major, self.ver_minor) >
                (other.ver_major, other.ver_minor))

    def __le__(self, other):
        return self < other or self == other

    def __ne__(self, other):
        return not self.__eq__(other)

    def __ge__(self, other):
        return self > other or self == other

    def matches(self, min_version, max_version):
        """Returns whether the version object represents a version
        greater than or equal to the minimum version and less than
        or equal to the maximum version.

        @param min_version: Minimum acceptable version.
        @param max_version: Maximum acceptable version.
        @returns: boolean

        If min_version is null then there is no minimum limit.
        If max_version is null then there is no maximum limit.
        If self is null then raise ValueError
        """

        if self.is_null():
            raise ValueError
        if max_version.is_null() and min_version.is_null():
            return True
        elif max_version.is_null():
            return min_version <= self
        elif min_version.is_null():
            return self <= max_version
        else:
            return min_version <= self <= max_version

    def get_string(self):
        """Converts object to string representation which if used to create
        an APIVersionRequest object results in the same version request.
        """
        if self.is_null():
            raise ValueError
        return "%s.%s" % (self.ver_major, self.ver_minor)


class VersionedMethod(object):

    def __init__(self, name, start_version, end_version, func):
        """Versioning information for a single method
        @name: Name of the method
        @start_version: Minimum acceptable version
        @end_version: Maximum acceptable_version
        @func: Method to call
        Minimum and maximums are inclusive
        """
        self.name = name
        self.start_version = start_version
        self.end_version = end_version
        self.func = func

    def __str__(self):
        return ("Version Method %s: min: %s, max: %s"
                % (self.name, self.start_version, self.end_version))
